@@ -21,9 +21,11 @@ package org.apache.sshd.client.session;
 
 import java.io.IOException;
 import java.net.SocketAddress;
-import java.nio.file.FileSystem;
+import java.nio.charset.Charset;
+import java.nio.charset.StandardCharsets;
 import java.security.KeyPair;
 import java.security.PublicKey;
+import java.util.Collections;
 import java.util.EnumMap;
 import java.util.List;
 import java.util.Map;
@@ -32,47 +34,52 @@ import java.util.concurrent.CopyOnWriteArrayList;
 
 import org.apache.sshd.client.ClientFactoryManager;
 import org.apache.sshd.client.auth.AuthenticationIdentitiesProvider;
-import org.apache.sshd.client.auth.UserAuth;
+import org.apache.sshd.client.auth.UserAuthFactory;
+import org.apache.sshd.client.auth.hostbased.HostBasedAuthenticationReporter;
 import org.apache.sshd.client.auth.keyboard.UserInteraction;
+import org.apache.sshd.client.auth.password.PasswordAuthenticationReporter;
 import org.apache.sshd.client.auth.password.PasswordIdentityProvider;
+import org.apache.sshd.client.auth.pubkey.PublicKeyAuthenticationReporter;
 import org.apache.sshd.client.channel.ChannelDirectTcpip;
 import org.apache.sshd.client.channel.ChannelExec;
 import org.apache.sshd.client.channel.ChannelShell;
 import org.apache.sshd.client.channel.ChannelSubsystem;
 import org.apache.sshd.client.channel.ClientChannel;
 import org.apache.sshd.client.keyverifier.ServerKeyVerifier;
-import org.apache.sshd.client.scp.DefaultScpClient;
-import org.apache.sshd.client.scp.ScpClient;
-import org.apache.sshd.client.subsystem.sftp.SftpClient;
-import org.apache.sshd.client.subsystem.sftp.SftpClientFactory;
-import org.apache.sshd.client.subsystem.sftp.SftpVersionSelector;
+import org.apache.sshd.common.AttributeRepository;
 import org.apache.sshd.common.FactoryManager;
-import org.apache.sshd.common.NamedFactory;
 import org.apache.sshd.common.NamedResource;
 import org.apache.sshd.common.RuntimeSshException;
+import org.apache.sshd.common.Service;
 import org.apache.sshd.common.SshConstants;
 import org.apache.sshd.common.SshException;
 import org.apache.sshd.common.channel.Channel;
+import org.apache.sshd.common.channel.PtyChannelConfigurationHolder;
 import org.apache.sshd.common.cipher.BuiltinCiphers;
 import org.apache.sshd.common.cipher.CipherNone;
 import org.apache.sshd.common.config.keys.KeyUtils;
-import org.apache.sshd.common.forward.ForwardingFilter;
+import org.apache.sshd.common.config.keys.OpenSshCertificate;
+import org.apache.sshd.common.forward.Forwarder;
 import org.apache.sshd.common.future.DefaultKeyExchangeFuture;
 import org.apache.sshd.common.future.KeyExchangeFuture;
 import org.apache.sshd.common.io.IoSession;
 import org.apache.sshd.common.io.IoWriteFuture;
 import org.apache.sshd.common.kex.KexProposalOption;
 import org.apache.sshd.common.kex.KexState;
-import org.apache.sshd.common.scp.ScpFileOpener;
-import org.apache.sshd.common.scp.ScpTransferEventListener;
+import org.apache.sshd.common.kex.extension.KexExtensionHandler;
+import org.apache.sshd.common.kex.extension.KexExtensionHandler.AvailabilityPhase;
+import org.apache.sshd.common.keyprovider.KeyIdentityProvider;
 import org.apache.sshd.common.session.ConnectionService;
-import org.apache.sshd.common.session.Session;
+import org.apache.sshd.common.session.SessionContext;
+import org.apache.sshd.common.session.SessionDisconnectHandler;
 import org.apache.sshd.common.session.helpers.AbstractConnectionService;
 import org.apache.sshd.common.session.helpers.AbstractSession;
+import org.apache.sshd.common.util.ExceptionUtils;
 import org.apache.sshd.common.util.GenericUtils;
 import org.apache.sshd.common.util.ValidateUtils;
 import org.apache.sshd.common.util.buffer.Buffer;
 import org.apache.sshd.common.util.net.SshdSocketAddress;
+import org.apache.sshd.core.CoreModuleProperties;
 
 /**
  * Provides default implementations of {@link ClientSession} related methods
@@ -80,21 +87,38 @@ import org.apache.sshd.common.util.net.SshdSocketAddress;
  * @author <a href="mailto:dev@mina.apache.org">Apache MINA SSHD Project</a>
  */
 public abstract class AbstractClientSession extends AbstractSession implements ClientSession {
+    protected final boolean sendImmediateClientIdentification;
+    protected final boolean sendImmediateKexInit;
+
     private final List<Object> identities = new CopyOnWriteArrayList<>();
     private final AuthenticationIdentitiesProvider identitiesProvider;
+    private final AttributeRepository connectionContext;
+
+    private PublicKey serverKey;
     private ServerKeyVerifier serverKeyVerifier;
     private UserInteraction userInteraction;
     private PasswordIdentityProvider passwordIdentityProvider;
-    private List<NamedFactory<UserAuth>> userAuthFactories;
-    private ScpTransferEventListener scpListener;
-    private ScpFileOpener scpOpener;
+    private PasswordAuthenticationReporter passwordAuthenticationReporter;
+    private KeyIdentityProvider keyIdentityProvider;
+    private PublicKeyAuthenticationReporter publicKeyAuthenticationReporter;
+    private HostBasedAuthenticationReporter hostBasedAuthenticationReporter;
+    private List<UserAuthFactory> userAuthFactories;
     private SocketAddress connectAddress;
     private ClientProxyConnector proxyConnector;
-    private SftpClientFactory sftpClientFactory;
 
     protected AbstractClientSession(ClientFactoryManager factoryManager, IoSession ioSession) {
         super(false, factoryManager, ioSession);
-        identitiesProvider = AuthenticationIdentitiesProvider.wrap(identities);
+
+        sendImmediateClientIdentification = CoreModuleProperties.SEND_IMMEDIATE_IDENTIFICATION.getRequired(this);
+        sendImmediateKexInit = CoreModuleProperties.SEND_IMMEDIATE_KEXINIT.getRequired(this);
+
+        identitiesProvider = AuthenticationIdentitiesProvider.wrapIdentities(identities);
+        connectionContext = (AttributeRepository) ioSession.getAttribute(AttributeRepository.class);
+    }
+
+    @Override
+    public AttributeRepository getConnectionContext() {
+        return connectionContext;
     }
 
     @Override
@@ -111,9 +135,24 @@ public abstract class AbstractClientSession extends AbstractSession implements C
         this.connectAddress = connectAddress;
     }
 
+    @Override
+    public PublicKey getServerKey() {
+        return serverKey;
+    }
+
+    public void setServerKey(PublicKey serverKey) {
+        if (log.isDebugEnabled()) {
+            log.debug("setServerKey({}) keyType={}, digest={}",
+                    this, KeyUtils.getKeyType(serverKey), KeyUtils.getFingerPrint(serverKey));
+        }
+
+        this.serverKey = serverKey;
+    }
+
     @Override
     public ServerKeyVerifier getServerKeyVerifier() {
-        return resolveEffectiveProvider(ServerKeyVerifier.class, serverKeyVerifier, getFactoryManager().getServerKeyVerifier());
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(ServerKeyVerifier.class, serverKeyVerifier, manager.getServerKeyVerifier());
     }
 
     @Override
@@ -123,7 +162,8 @@ public abstract class AbstractClientSession extends AbstractSession implements C
 
     @Override
     public UserInteraction getUserInteraction() {
-        return resolveEffectiveProvider(UserInteraction.class, userInteraction, getFactoryManager().getUserInteraction());
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(UserInteraction.class, userInteraction, manager.getUserInteraction());
     }
 
     @Override
@@ -132,12 +172,25 @@ public abstract class AbstractClientSession extends AbstractSession implements C
     }
 
     @Override
-    public List<NamedFactory<UserAuth>> getUserAuthFactories() {
-        return resolveEffectiveFactories(UserAuth.class, userAuthFactories, getFactoryManager().getUserAuthFactories());
+    public PasswordAuthenticationReporter getPasswordAuthenticationReporter() {
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(PasswordAuthenticationReporter.class, passwordAuthenticationReporter,
+                manager.getPasswordAuthenticationReporter());
     }
 
     @Override
-    public void setUserAuthFactories(List<NamedFactory<UserAuth>> userAuthFactories) {
+    public void setPasswordAuthenticationReporter(PasswordAuthenticationReporter reporter) {
+        this.passwordAuthenticationReporter = reporter;
+    }
+
+    @Override
+    public List<UserAuthFactory> getUserAuthFactories() {
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveFactories(userAuthFactories, manager.getUserAuthFactories());
+    }
+
+    @Override
+    public void setUserAuthFactories(List<UserAuthFactory> userAuthFactories) {
         this.userAuthFactories = userAuthFactories; // OK if null/empty - inherit from parent
     }
 
@@ -148,7 +201,9 @@ public abstract class AbstractClientSession extends AbstractSession implements C
 
     @Override
     public PasswordIdentityProvider getPasswordIdentityProvider() {
-        return resolveEffectiveProvider(PasswordIdentityProvider.class, passwordIdentityProvider, getFactoryManager().getPasswordIdentityProvider());
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(PasswordIdentityProvider.class, passwordIdentityProvider,
+                manager.getPasswordIdentityProvider());
     }
 
     @Override
@@ -156,9 +211,46 @@ public abstract class AbstractClientSession extends AbstractSession implements C
         passwordIdentityProvider = provider;
     }
 
+    @Override
+    public KeyIdentityProvider getKeyIdentityProvider() {
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(KeyIdentityProvider.class, keyIdentityProvider,
+                manager.getKeyIdentityProvider());
+    }
+
+    @Override
+    public void setKeyIdentityProvider(KeyIdentityProvider keyIdentityProvider) {
+        this.keyIdentityProvider = keyIdentityProvider;
+    }
+
+    @Override
+    public PublicKeyAuthenticationReporter getPublicKeyAuthenticationReporter() {
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(PublicKeyAuthenticationReporter.class, publicKeyAuthenticationReporter,
+                manager.getPublicKeyAuthenticationReporter());
+    }
+
+    @Override
+    public void setPublicKeyAuthenticationReporter(PublicKeyAuthenticationReporter reporter) {
+        this.publicKeyAuthenticationReporter = reporter;
+    }
+
+    @Override
+    public HostBasedAuthenticationReporter getHostBasedAuthenticationReporter() {
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(HostBasedAuthenticationReporter.class, hostBasedAuthenticationReporter,
+                manager.getHostBasedAuthenticationReporter());
+    }
+
+    @Override
+    public void setHostBasedAuthenticationReporter(HostBasedAuthenticationReporter reporter) {
+        this.hostBasedAuthenticationReporter = reporter;
+    }
+
     @Override
     public ClientProxyConnector getClientProxyConnector() {
-        return resolveEffectiveProvider(ClientProxyConnector.class, proxyConnector, getFactoryManager().getClientProxyConnector());
+        ClientFactoryManager manager = getFactoryManager();
+        return resolveEffectiveProvider(ClientProxyConnector.class, proxyConnector, manager.getClientProxyConnector());
     }
 
     @Override
@@ -166,16 +258,6 @@ public abstract class AbstractClientSession extends AbstractSession implements C
         this.proxyConnector = proxyConnector;
     }
 
-    @Override
-    public SftpClientFactory getSftpClientFactory() {
-        return resolveEffectiveProvider(SftpClientFactory.class, sftpClientFactory, getFactoryManager().getSftpClientFactory());
-    }
-
-    @Override
-    public void setSftpClientFactory(SftpClientFactory sftpClientFactory) {
-        this.sftpClientFactory = sftpClientFactory;
-    }
-
     @Override
     public void addPasswordIdentity(String password) {
         // DO NOT USE checkNotNullOrNotEmpty SINCE IT TRIMS THE RESULT
@@ -192,8 +274,8 @@ public abstract class AbstractClientSession extends AbstractSession implements C
             return null;
         }
 
-        int index = AuthenticationIdentitiesProvider.findIdentityIndex(
-                identities, AuthenticationIdentitiesProvider.PASSWORD_IDENTITY_COMPARATOR, password);
+        int index = AuthenticationIdentitiesProvider.findIdentityIndex(identities,
+                AuthenticationIdentitiesProvider.PASSWORD_IDENTITY_COMPARATOR, password);
         if (index >= 0) {
             return (String) identities.remove(index);
         } else {
@@ -211,8 +293,7 @@ public abstract class AbstractClientSession extends AbstractSession implements C
 
         if (log.isDebugEnabled()) {
             PublicKey key = kp.getPublic();
-            log.debug("addPublicKeyIdentity({}) {}-{}",
-                      this, KeyUtils.getKeyType(key), KeyUtils.getFingerPrint(key));
+            log.debug("addPublicKeyIdentity({}) {}-{}", this, KeyUtils.getKeyType(key), KeyUtils.getFingerPrint(key));
         }
     }
 
@@ -222,8 +303,8 @@ public abstract class AbstractClientSession extends AbstractSession implements C
             return null;
         }
 
-        int index = AuthenticationIdentitiesProvider.findIdentityIndex(
-                identities, AuthenticationIdentitiesProvider.KEYPAIR_IDENTITY_COMPARATOR, kp);
+        int index = AuthenticationIdentitiesProvider.findIdentityIndex(identities,
+                AuthenticationIdentitiesProvider.KEYPAIR_IDENTITY_COMPARATOR, kp);
         if (index >= 0) {
             return (KeyPair) identities.remove(index);
         } else {
@@ -231,29 +312,55 @@ public abstract class AbstractClientSession extends AbstractSession implements C
         }
     }
 
-    protected IoWriteFuture sendClientIdentification() throws Exception {
-        clientVersion = resolveIdentificationString(ClientFactoryManager.CLIENT_IDENTIFICATION);
-
-        ClientProxyConnector proxyConnector = getClientProxyConnector();
-        if (proxyConnector != null) {
-            try {
-                proxyConnector.sendClientProxyMetadata(this);
-            } catch (Throwable t) {
-                log.warn("sendClientIdentification({}) failed ({}) to send proxy metadata: {}",
-                         this, t.getClass().getSimpleName(), t.getMessage());
-                if (log.isDebugEnabled()) {
-                    log.debug("sendClientIdentification(" + this + ") proxy metadata send failure details", t);
-                }
-
-                if (t instanceof Exception) {
-                    throw (Exception) t;
-                } else {
-                    throw new RuntimeSshException(t);
-                }
+    protected void initializeKeyExchangePhase() throws Exception {
+        KexExtensionHandler extHandler = getKexExtensionHandler();
+        if ((extHandler == null) || (!extHandler.isKexExtensionsAvailable(this, AvailabilityPhase.PREKEX))) {
+            kexState.set(KexState.INIT);
+            sendKexInit();
+        } else {
+            if (log.isDebugEnabled()) {
+                log.debug("initializeKexPhase({}) delay KEX-INIT until server-side one received", this);
             }
         }
+    }
 
-        return sendIdentification(clientVersion);
+    protected void initializeProxyConnector() throws Exception {
+        ClientProxyConnector proxyConnector = getClientProxyConnector();
+        boolean debugEnabled = log.isDebugEnabled();
+        if (proxyConnector == null) {
+            if (debugEnabled) {
+                log.debug("initializeProxyConnector({}) no proxy to initialize", this);
+            }
+            return;
+        }
+
+        try {
+            if (debugEnabled) {
+                log.debug("initializeProxyConnector({}) initialize proxy={}", this, proxyConnector);
+            }
+
+            proxyConnector.sendClientProxyMetadata(this);
+
+            if (debugEnabled) {
+                log.debug("initializeProxyConnector({}) proxy={} initialized", this, proxyConnector);
+            }
+        } catch (Throwable t) {
+            warn("initializeProxyConnector({}) failed ({}) to send proxy metadata: {}",
+                    this, t.getClass().getSimpleName(), t.getMessage(), t);
+
+            if (t instanceof Exception) {
+                throw (Exception) t;
+            } else {
+                throw new RuntimeSshException(t);
+            }
+        }
+    }
+
+    protected IoWriteFuture sendClientIdentification() throws Exception {
+        clientVersion = resolveIdentificationString(CoreModuleProperties.CLIENT_IDENTIFICATION.getName());
+        // Note: we intentionally use an unmodifiable list in order to enforce the fact that client cannot send header lines
+        signalSendIdentification(clientVersion, Collections.emptyList());
+        return sendIdentification(clientVersion, Collections.emptyList());
     }
 
     @Override
@@ -270,17 +377,32 @@ public abstract class AbstractClientSession extends AbstractSession implements C
         } else if (Channel.CHANNEL_SUBSYSTEM.equals(type)) {
             return createSubsystemChannel(subType);
         } else {
-            throw new IllegalArgumentException("Unsupported channel type " + type);
+            throw new IllegalArgumentException("Unsupported channel type requested: " + type);
         }
     }
 
     @Override
-    public ChannelExec createExecChannel(String command) throws IOException {
-        ChannelExec channel = new ChannelExec(command);
+    public ChannelExec createExecChannel(
+            String command, Charset charset, PtyChannelConfigurationHolder ptyConfig,
+            Map<String, ?> env) throws IOException {
+        ChannelExec channel = new ChannelExec(command, charset, ptyConfig, env);
         ConnectionService service = getConnectionService();
-        int id = service.registerChannel(channel);
+        long id = service.registerChannel(channel);
         if (log.isDebugEnabled()) {
-            log.debug("createExecChannel({})[{}] created id={}", this, command, id);
+            log.debug("createExecChannel({})[{}] created id={} - PTY={}", this, command, id, ptyConfig);
+        }
+        return channel;
+    }
+
+    @Override
+    public ChannelExec createExecChannel(byte[] command, PtyChannelConfigurationHolder ptyConfig, Map<String, ?> env)
+            throws IOException {
+        ChannelExec channel = new ChannelExec(command, ptyConfig, env);
+        ConnectionService service = getConnectionService();
+        long id = service.registerChannel(channel);
+        if (log.isDebugEnabled()) {
+            log.debug("createExecChannel({})[{}] created id={} - PTY={}", this, new String(command, StandardCharsets.UTF_8), id,
+                    ptyConfig);
         }
         return channel;
     }
@@ -289,7 +411,7 @@ public abstract class AbstractClientSession extends AbstractSession implements C
     public ChannelSubsystem createSubsystemChannel(String subsystem) throws IOException {
         ChannelSubsystem channel = new ChannelSubsystem(subsystem);
         ConnectionService service = getConnectionService();
-        int id = service.registerChannel(channel);
+        long id = service.registerChannel(channel);
         if (log.isDebugEnabled()) {
             log.debug("createSubsystemChannel({})[{}] created id={}", this, subsystem, id);
         }
@@ -297,10 +419,11 @@ public abstract class AbstractClientSession extends AbstractSession implements C
     }
 
     @Override
-    public ChannelDirectTcpip createDirectTcpipChannel(SshdSocketAddress local, SshdSocketAddress remote) throws IOException {
+    public ChannelDirectTcpip createDirectTcpipChannel(SshdSocketAddress local, SshdSocketAddress remote)
+            throws IOException {
         ChannelDirectTcpip channel = new ChannelDirectTcpip(local, remote);
         ConnectionService service = getConnectionService();
-        int id = service.registerChannel(channel);
+        long id = service.registerChannel(channel);
         if (log.isDebugEnabled()) {
             log.debug("createDirectTcpipChannel({})[{} => {}] created id={}", this, local, remote, id);
         }
@@ -311,86 +434,53 @@ public abstract class AbstractClientSession extends AbstractSession implements C
         return getService(ClientUserAuthService.class);
     }
 
+    @Override
     protected ConnectionService getConnectionService() {
         return getService(ConnectionService.class);
     }
 
     @Override
-    public ScpFileOpener getScpFileOpener() {
-        return resolveEffectiveProvider(ScpFileOpener.class, scpOpener, getFactoryManager().getScpFileOpener());
-    }
-
-    @Override
-    public void setScpFileOpener(ScpFileOpener opener) {
-        scpOpener = opener;
-    }
-
-    @Override
-    public ScpTransferEventListener getScpTransferEventListener() {
-        return scpListener;
-    }
-
-    @Override
-    public void setScpTransferEventListener(ScpTransferEventListener listener) {
-        scpListener = listener;
-    }
-
-    @Override
-    public ScpClient createScpClient(ScpFileOpener opener, ScpTransferEventListener listener) {
-        return new DefaultScpClient(this, opener, listener);
-    }
-
-    @Override
-    public SftpClient createSftpClient(SftpVersionSelector selector) throws IOException {
-        SftpClientFactory factory = getSftpClientFactory();
-        return factory.createSftpClient(this, selector);
-    }
-
-    @Override
-    public FileSystem createSftpFileSystem(SftpVersionSelector selector, int readBufferSize, int writeBufferSize) throws IOException {
-        SftpClientFactory factory = getSftpClientFactory();
-        return factory.createSftpFileSystem(this, selector, readBufferSize, writeBufferSize);
-    }
-
-    @Override
-    public SshdSocketAddress startLocalPortForwarding(SshdSocketAddress local, SshdSocketAddress remote) throws IOException {
-        ForwardingFilter filter = getForwardingFilter();
-        return filter.startLocalPortForwarding(local, remote);
+    public SshdSocketAddress startLocalPortForwarding(SshdSocketAddress local, SshdSocketAddress remote)
+            throws IOException {
+        Forwarder forwarder = getForwarder();
+        return forwarder.startLocalPortForwarding(local, remote);
     }
 
     @Override
     public void stopLocalPortForwarding(SshdSocketAddress local) throws IOException {
-        ForwardingFilter filter = getForwardingFilter();
-        filter.stopLocalPortForwarding(local);
+        Forwarder forwarder = getForwarder();
+        forwarder.stopLocalPortForwarding(local);
     }
 
     @Override
-    public SshdSocketAddress startRemotePortForwarding(SshdSocketAddress remote, SshdSocketAddress local) throws IOException {
-        ForwardingFilter filter = getForwardingFilter();
-        return filter.startRemotePortForwarding(remote, local);
+    public SshdSocketAddress startRemotePortForwarding(SshdSocketAddress remote, SshdSocketAddress local)
+            throws IOException {
+        Forwarder forwarder = getForwarder();
+        return forwarder.startRemotePortForwarding(remote, local);
     }
 
     @Override
     public void stopRemotePortForwarding(SshdSocketAddress remote) throws IOException {
-        ForwardingFilter filter = getForwardingFilter();
-        filter.stopRemotePortForwarding(remote);
+        Forwarder forwarder = getForwarder();
+        forwarder.stopRemotePortForwarding(remote);
     }
 
     @Override
     public SshdSocketAddress startDynamicPortForwarding(SshdSocketAddress local) throws IOException {
-        ForwardingFilter filter = getForwardingFilter();
-        return filter.startDynamicPortForwarding(local);
+        Forwarder forwarder = getForwarder();
+        return forwarder.startDynamicPortForwarding(local);
     }
 
     @Override
     public void stopDynamicPortForwarding(SshdSocketAddress local) throws IOException {
-        ForwardingFilter filter = getForwardingFilter();
-        filter.stopDynamicPortForwarding(local);
+        Forwarder forwarder = getForwarder();
+        forwarder.stopDynamicPortForwarding(local);
     }
 
-    protected ForwardingFilter getForwardingFilter() {
+    @Override
+    protected Forwarder getForwarder() {
         ConnectionService service = Objects.requireNonNull(getConnectionService(), "No connection service");
-        return Objects.requireNonNull(service.getForwardingFilter(), "No forwarder");
+        return Objects.requireNonNull(service.getForwarder(), "No forwarder");
     }
 
     @Override
@@ -401,27 +491,37 @@ public abstract class AbstractClientSession extends AbstractSession implements C
     }
 
     @Override
-    public void startService(String name) throws Exception {
+    public void startService(String name, Buffer buffer) throws Exception {
+        SessionDisconnectHandler handler = getSessionDisconnectHandler();
+        if ((handler != null) && handler.handleUnsupportedServiceDisconnectReason(this,
+                SshConstants.SSH_MSG_SERVICE_REQUEST, name, buffer)) {
+            if (log.isDebugEnabled()) {
+                log.debug("startService({}) ignore unknown service={} by handler", this, name);
+            }
+            return;
+        }
+
         throw new IllegalStateException("Starting services is not supported on the client side: " + name);
     }
 
     @Override
-    public ChannelShell createShellChannel() throws IOException {
+    public ChannelShell createShellChannel(PtyChannelConfigurationHolder ptyConfig, Map<String, ?> env)
+            throws IOException {
         if ((inCipher instanceof CipherNone) || (outCipher instanceof CipherNone)) {
             throw new IllegalStateException("Interactive channels are not supported with none cipher");
         }
 
-        ChannelShell channel = new ChannelShell();
+        ChannelShell channel = new ChannelShell(ptyConfig, env);
         ConnectionService service = getConnectionService();
-        int id = service.registerChannel(channel);
+        long id = service.registerChannel(channel);
         if (log.isDebugEnabled()) {
-            log.debug("createShellChannel({}) created id={}", this, id);
+            log.debug("createShellChannel({}) created id={} - PTY={}", this, id, ptyConfig);
         }
         return channel;
     }
 
     @Override
-    protected boolean readIdentification(Buffer buffer) throws IOException {
+    protected boolean readIdentification(Buffer buffer) throws Exception {
         List<String> ident = doReadIdentification(buffer, false);
         int numLines = GenericUtils.size(ident);
         serverVersion = (numLines <= 0) ? null : ident.remove(numLines - 1);
@@ -433,16 +533,29 @@ public abstract class AbstractClientSession extends AbstractSession implements C
             log.debug("readIdentification({}) Server version string: {}", this, serverVersion);
         }
 
-        if (!Session.isValidVersionPrefix(serverVersion)) {
+        if (!SessionContext.isValidVersionPrefix(serverVersion)) {
             throw new SshException(SshConstants.SSH2_DISCONNECT_PROTOCOL_VERSION_NOT_SUPPORTED,
                     "Unsupported protocol version: " + serverVersion);
         }
 
-        signalExtraServerVersionInfo(ident);
+        signalExtraServerVersionInfo(serverVersion, ident);
+
+        // Now that we have the server's identity reported see if have delayed any of our duties...
+        if (!sendImmediateClientIdentification) {
+            sendClientIdentification();
+            // if client identification not sent then KEX-INIT was not sent either
+            initializeKeyExchangePhase();
+        } else if (!sendImmediateKexInit) {
+            // if client identification sent, perhaps we delayed KEX-INIT
+            initializeKeyExchangePhase();
+        }
+
         return true;
     }
 
-    protected void signalExtraServerVersionInfo(List<String> lines) throws IOException {
+    protected void signalExtraServerVersionInfo(String version, List<String> lines) throws Exception {
+        signalPeerIdentificationReceived(version, lines);
+
         if (GenericUtils.isEmpty(lines)) {
             return;
         }
@@ -453,42 +566,79 @@ public abstract class AbstractClientSession extends AbstractSession implements C
                 ui.serverVersionInfo(this, lines);
             }
         } catch (Error e) {
-            log.warn("signalExtraServerVersionInfo({}) failed ({}) to consult interaction: {}",
-                     this, e.getClass().getSimpleName(), e.getMessage());
-            if (log.isDebugEnabled()) {
-                log.debug("signalExtraServerVersionInfo(" + this + ") interaction consultation failure details", e);
-            }
-
+            warn("signalExtraServerVersionInfo({})[{}] failed ({}) to consult interaction: {}",
+                    this, version, e.getClass().getSimpleName(), e.getMessage(), e);
             throw new RuntimeSshException(e);
         }
     }
 
     @Override
-    protected byte[] sendKexInit(Map<KexProposalOption, String> proposal) throws IOException {
+    protected byte[] sendKexInit(Map<KexProposalOption, String> proposal) throws Exception {
         mergeProposals(clientProposal, proposal);
         return super.sendKexInit(proposal);
     }
 
     @Override
     protected void setKexSeed(byte... seed) {
-        i_c = ValidateUtils.checkNotNullAndNotEmpty(seed, "No KEX seed");
+        setClientKexData(seed);
+    }
+
+    @Override
+    protected byte[] receiveKexInit(Buffer buffer) throws Exception {
+        byte[] seed = super.receiveKexInit(buffer);
+        /*
+         * Check if the session has delayed its KEX-INIT until the server's one was received in order to support KEX
+         * extension negotiation (RFC 8308).
+         */
+        if (kexState.compareAndSet(KexState.UNKNOWN, KexState.RUN)) {
+            if (log.isDebugEnabled()) {
+                log.debug("receiveKexInit({}) sending client proposal", this);
+            }
+            kexState.set(KexState.INIT);
+            sendKexInit();
+        }
+
+        return seed;
     }
 
     @Override
     protected void receiveKexInit(Map<KexProposalOption, String> proposal, byte[] seed) throws IOException {
         mergeProposals(serverProposal, proposal);
-        i_s = seed;
+        setServerKexData(seed);
     }
 
     @Override
-    protected void checkKeys() throws SshException {
+    protected void checkKeys() throws IOException {
         ServerKeyVerifier serverKeyVerifier = Objects.requireNonNull(getServerKeyVerifier(), "No server key verifier");
-        SocketAddress remoteAddress = ioSession.getRemoteAddress();
-        PublicKey serverKey = kex.getServerKey();
-        boolean verified = serverKeyVerifier.verifyServerKey(this, remoteAddress, serverKey);
-        if (log.isDebugEnabled()) {
-            log.debug("checkKeys({}) key={}-{}, verified={}",
-                      this, KeyUtils.getKeyType(serverKey), KeyUtils.getFingerPrint(serverKey), verified);
+        IoSession networkSession = getIoSession();
+        SocketAddress remoteAddress = networkSession.getRemoteAddress();
+        PublicKey serverKey = Objects.requireNonNull(getServerKey(), "No server key to verify");
+        SshdSocketAddress targetServerAddress = getAttribute(ClientSessionCreator.TARGET_SERVER);
+        if (targetServerAddress != null) {
+            remoteAddress = targetServerAddress.toInetSocketAddress();
+        }
+
+        boolean verified = false;
+        if (serverKey instanceof OpenSshCertificate) {
+            // check if we trust the CA
+            verified = serverKeyVerifier.verifyServerKey(this, remoteAddress, ((OpenSshCertificate) serverKey).getCaPubKey());
+            if (log.isDebugEnabled()) {
+                log.debug("checkCA({}) key={}-{}, verified={}",
+                        this, KeyUtils.getKeyType(serverKey), KeyUtils.getFingerPrint(serverKey), verified);
+            }
+
+            if (!verified) {
+                // fallback to actual public host key
+                serverKey = ((OpenSshCertificate) serverKey).getCertPubKey();
+            }
+        }
+
+        if (!verified) {
+            verified = serverKeyVerifier.verifyServerKey(this, remoteAddress, serverKey);
+            if (log.isDebugEnabled()) {
+                log.debug("checkKeys({}) key={}-{}, verified={}",
+                        this, KeyUtils.getKeyType(serverKey), KeyUtils.getFingerPrint(serverKey), verified);
+            }
         }
 
         if (!verified) {
@@ -498,8 +648,9 @@ public abstract class AbstractClientSession extends AbstractSession implements C
 
     @Override
     public KeyExchangeFuture switchToNoneCipher() throws IOException {
-        if (!(currentService instanceof AbstractConnectionService<?>)
-                || !GenericUtils.isEmpty(((AbstractConnectionService<?>) currentService).getChannels())) {
+        Service service = currentService.getService();
+        if (!(service instanceof AbstractConnectionService)
+                || !GenericUtils.isEmpty(((AbstractConnectionService) service).getChannels())) {
             throw new IllegalStateException("The switch to the none cipher must be done immediately after authentication");
         }
 
@@ -507,12 +658,7 @@ public abstract class AbstractClientSession extends AbstractSession implements C
             DefaultKeyExchangeFuture kexFuture = new DefaultKeyExchangeFuture(toString(), null);
             DefaultKeyExchangeFuture prev = kexFutureHolder.getAndSet(kexFuture);
             if (prev != null) {
-                synchronized (prev) {
-                    Object value = prev.getValue();
-                    if (value == null) {
-                        prev.setValue(new SshException("Switch to none cipher while previous KEX is ongoing"));
-                    }
-                }
+                prev.setValue(new SshException("Switch to none cipher while previous KEX is ongoing"));
             }
 
             String c2sEncServer;
@@ -549,8 +695,21 @@ public abstract class AbstractClientSession extends AbstractSession implements C
                 proposal.put(KexProposalOption.C2SENC, BuiltinCiphers.Constants.NONE);
                 proposal.put(KexProposalOption.S2CENC, BuiltinCiphers.Constants.NONE);
 
-                byte[] seed = sendKexInit(proposal);
-                setKexSeed(seed);
+                synchronized (kexState) {
+                    DefaultKeyExchangeFuture initFuture = kexInitializedFuture;
+                    if (initFuture == null) {
+                        initFuture = new DefaultKeyExchangeFuture(toString(), null);
+                        kexInitializedFuture = initFuture;
+                    }
+                    try {
+                        byte[] seed = sendKexInit(proposal);
+                        setKexSeed(seed);
+                        initFuture.setValue(Boolean.TRUE);
+                    } catch (Exception e) {
+                        initFuture.setValue(e);
+                        ExceptionUtils.rethrowAsIoException(e);
+                    }
+                }
             }
 
             return Objects.requireNonNull(kexFutureHolder.get(), "No current KEX future");
